import logging
import os
import sys
import time
import multiprocessing
import pathlib
import subprocess

import psutil
import servicemanager
import win32event
import win32service
import win32serviceutil

# service constants for watchdog operation
SSA_TIME_TO_RUN = os.getenv("SSA_TIME_TO_RUN", 6)  # sec
if int(SSA_TIME_TO_RUN) < 1:
    raise ValueError("SSA_TIME_TO_RUN must be int and more 0.")
SSA_RETRY_DELAY = os.getenv("SSA_RETRY_DELAY", 3)  # sec
if int(SSA_TIME_TO_RUN) < 1:
    raise ValueError("SSA_RETRY_DELAY must be int and more 0.")
SSA_RETRIES_COUNT = os.getenv("SSA_RETRIES_COUNT", 3)
if int(SSA_RETRIES_COUNT) < 0:
    raise ValueError("SSA_RETRIES_COUNT must be integer and non-negative.")


# server startup constants
SSA_LISTEN = os.getenv("SSA_LISTEN", "0.0.0.0")
SSA_PORT = os.getenv("SSA_PORT", 31000)
SSA_SERVICE_NAME = os.getenv("SSA_NAME", "TermideskSessionAgent.exe")


def setup_logging():
    """
    Logger setup

    Returns:
        logging: logger instance
    """

    PROGRAM_DATA = os.path.expandvars(r"%PROGRAMDATA%")
    LOG_FILE_PATH = os.path.join(
        PROGRAM_DATA,
        "UVEON",
        "Termidesk Session Agent",
        "log",
        "session-agent-service.log",
    )
    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s:%(levelname)s:%(name)s:%(message)s",
        filename=LOG_FILE_PATH,
        filemode="a",
    )


class Watchdog:
    """
    Class for monitoring the API server process
    """

    def __init__(self):

        self.watch_wait_time = SSA_RETRY_DELAY
        self.time_to_run = SSA_TIME_TO_RUN
        self.max_tryes_restart = SSA_RETRIES_COUNT
        self.count_of_fails = 0

        self.keep_running = True
        self.ready_to_stop = False
        self.process = None
        self.process_name = SSA_SERVICE_NAME
        self.child_pid = None

    def run(self):
        """
        Process state monitoring method
        """

        logger.info("Watchdog running")

        while self.keep_running:
            if not self.check_ssa_processes():
                self.count_of_fails += 1
                if self.count_of_fails > self.max_tryes_restart:
                    # report startup error more than 3 times
                    logger.critical(
                        "Watchdog: The number of attempts to start the service exceeds "
                        f"{self.max_tryes_restart}. Terminating."
                    )
                    # Finishing the handler
                    self.terminate()
                    self.kill_proc_tree()
                    sys.exit(0)
                # report fault
                logger.error(
                    f"Watchdog: target process is down. Restart attempt {self.count_of_fails}."
                )
                # restart the target process
                self.boot_process()

            logger.debug(f"Processes continues to run. PIDs > {self.get_proc_pids()}")
            # block for a moment
            time.sleep(self.watch_wait_time)

    def boot_process(self):
        """
        Creates and starts a server API processes.

        Returns:
            Process instance
        """
        # Todo: для данной реализации необходимо проанализировать и переработать обработку
        #  родительских/дочерних процессов враппером. В текущем виде получаем ошибку.
        # app_path = f"{pathlib.Path(__file__).parent.resolve()}\\app.py"
        # self.process = subprocess.Popen(
        #     [sys.executable, app_path],
        #     stdout=subprocess.PIPE,
        #     stderr=subprocess.PIPE,
        #     bufsize=subprocess.PIPE,
        # )

        server_path = os.path.join(sys._MEIPASS, "server\\TermideskSessionAgent.exe")
        server_cmd = f"call {server_path}"
        self.process = subprocess.Popen(server_cmd, shell=True)

        time.sleep(self.time_to_run)

        return self.process

    def check_ssa_processes(self):
        """
        Method checks the state of the session agent processes.

        Returns:
            bool: Flag indicating whether the ssa processes are alive
        """
        processes_exists = any(self.get_proc_pids())

        return processes_exists

    def get_proc_pids(self):
        """Getting the list of active session agent processes PIDs"""
        ssa_processes = []
        for proc in psutil.process_iter():
            if self.process_name in proc.name():
                ssa_processes.append(psutil.Process(proc.pid))

        return ssa_processes

    def kill_proc_tree(self):
        """SSA processes termination method"""
        for proc in self.get_proc_pids():
            proc.kill()

        self.ready_to_stop = True

    def terminate(self):
        """
        Termination method of SSA processes monitoring
        """

        self.keep_running = False


class SessionAgentService(win32serviceutil.ServiceFramework):
    _svc_name_ = "TermideskSessionAgentService"
    _svc_display_name_ = "Termidesk Session Agent Service"
    _svc_description_ = "Session Agent Service for communication with terminal service."

    def __init__(self, args):
        win32serviceutil.ServiceFramework.__init__(self, args)
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)

    def SvcStop(self):
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)

        self.watchdog.terminate()
        self.watchdog.kill_proc_tree()
        logger.info("Watchdog was stoped by a STOP signal .")

        win32event.SetEvent(self.hWaitStop)
        self.ReportServiceStatus(win32service.SERVICE_STOPPED)

    def SvcDoRun(self):
        self.watchdog = Watchdog()
        self.watchdog.boot_process()
        self.watchdog.run()

        while True:
            win32event.WaitForSingleObject(self.hWaitStop, 0)
            if self.watchdog.ready_to_stop:
                self.SvcStop()


def main():
    """
    Start wrapper
    """

    # Needed for pyinstaller for multiprocessing on WindowsOS
    multiprocessing.freeze_support()
    try:
        if len(sys.argv) == 1:
            servicemanager.Initialize()
            servicemanager.PrepareToHostSingle(SessionAgentService)
            servicemanager.StartServiceCtrlDispatcher()
        else:
            win32serviceutil.HandleCommandLine(SessionAgentService)
    except Exception as ex:
        logger.error(ex)
        raise Exception from ex


if __name__ == "__main__":
    setup_logging()
    logger = logging.getLogger(__name__)
    main()
