import json
import logging
import traceback
from typing import Any, Optional, AsyncGenerator, Dict, Union
from uuid import uuid4

import anyio
from fastapi import Request
from fastapi.responses import JSONResponse
from starlette.datastructures import MutableHeaders
from starlette.middleware.base import (
    BaseHTTPMiddleware,
    DispatchFunction,
    RequestResponseEndpoint,
)
from starlette.responses import PlainTextResponse, Response, StreamingResponse

from starlette.status import (
    HTTP_500_INTERNAL_SERVER_ERROR,
    HTTP_501_NOT_IMPLEMENTED,
    HTTP_401_UNAUTHORIZED,
)

from starlette.types import ASGIApp, Receive, Scope, Send

from constants.common_const import ConstMiddleware as const
from excepts import (
    AuthenticationError,
    BaseSSAException,
)

logger = logging.getLogger("middleware.request")


async def log_request_body(request: Request) -> None:
    """Logging request body: FastAPI dependency"""
    try:
        req_body = (await request.json()).copy()
        logger.debug("[{0}] Request body: {1}".format(
            request.headers.get("request_id"),
            json_secure(req_body)
            ))

    except json.decoder.JSONDecodeError:
        pass


def json_secure(body: Dict[Any, Any]) -> Dict[Any, Any]:
    """Replace value in secure key on SECURE_VALUE_MASK"""
    for field in const.SECURE_FIELDS:
        if field in body:
            body[field] = const.SECURE_VALUE_MASK
    return body


class BaseLoggingMiddleware(BaseHTTPMiddleware):
    """Base logging middleware, added log message after route info message"""

    def __init__(
        self, app: ASGIApp, dispatch: Optional[DispatchFunction] = None
    ) -> None:
        super().__init__(app, dispatch)
        self.__queue_log = []

    def add_to_logging(self, log_lvl: int, msg: str):
        """Adds a func logging to the queue"""

        def _inner():
            logger.log(log_lvl, msg)

        self.__queue_log.append(_inner)

    def safe_logging(self):
        """Print logging message from queue"""
        for func in self.__queue_log:
            func()

        self.__queue_log = []

    def error_response(self, request: Request, exc: Exception) -> Response:
        """Make error response, code from ServerErrorMiddleware"""
        return PlainTextResponse("Internal Server Error", status_code=500)

    def _add_request_id_header(self, scope: Scope) -> None:
        self.request_id = str(uuid4().fields[-1])[:5]
        scope["headers"].append((b"request_id", f"{self.request_id}".encode()))

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:  # noqa: C901
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        self._add_request_id_header(scope=scope)

        async def call_next(request: Request) -> Response:
            app_exc: Optional[Exception] = None
            send_stream, recv_stream = anyio.create_memory_object_stream()

            async def coro() -> None:
                nonlocal app_exc

                async with send_stream:
                    try:
                        await self.app(scope, request.receive, send_stream.send)
                    except Exception as exc:
                        app_exc = exc

            task_group.start_soon(coro)

            try:
                message = await recv_stream.receive()
            except anyio.EndOfStream:
                if app_exc is not None:

                    request = Request(scope)  # from ServerErrorMiddleware
                    response = self.error_response(request, app_exc)
                    await response(scope, receive, send)

                    self.safe_logging()

                    raise app_exc
                raise RuntimeError("No response returned.")

            assert message["type"] == "http.response.start"

            async def body_stream() -> AsyncGenerator[bytes, None]:
                async with recv_stream:
                    async for message in recv_stream:
                        assert message["type"] == "http.response.body"
                        yield message.get("body", b"")

                if app_exc is not None:
                    raise app_exc

            response = StreamingResponse(
                status_code=message["status"], content=body_stream()
            )
            response.raw_headers = message["headers"]
            return response

        async with anyio.create_task_group() as task_group:
            request = Request(scope, receive=receive)
            response = await self.dispatch_func(request, call_next)
            await response(scope, receive, send)
            self.safe_logging()
            task_group.cancel_scope.cancel()


class LoggedBodyMiddleware(BaseLoggingMiddleware):
    @staticmethod
    async def get_response_body(response: Response) -> Dict[Any, Any]:
        """Return response body"""
        res_body = b"".join([chunk async for chunk in response.body_iterator])
        return json.loads(res_body)

    @staticmethod
    def _request_to_log(session_uuid: str, request: Request) -> str:
        """Form string of params based on request to log"""
        host, port = request.client.host, request.client.port
        method = request.method
        path = request.url.path
        scheme = request.url.scheme.upper()

        return f'[{session_uuid}] {host}:{port} - "{method} {path}" [{scheme}]'

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Response:
        # Log incoming request
        logger.info(self._request_to_log(self.request_id, request))

        response = await call_next(request)

        self.add_to_logging(
            logging.INFO,
            f'[{self.request_id}] Status code: {response.status_code}'
        )

        if self.is_response_body_json(response):
            json_res_body = await self.get_response_body(response)

            if json_res_body.get("traceback"):
                traceback_ = json_res_body.pop("traceback")
                exc_level = json_res_body.pop("exc_level")
                exc_detail = json_res_body.get("detail")
                # Логирование трассировки только для исключений уровня ERROR и выше
                if exc_level >= logging.ERROR:
                    self.add_to_logging(exc_level, traceback_)
                else:
                    self.add_to_logging(exc_level, exc_detail)
                    # Временное решение для логирования трассировки
                    self.add_to_logging(logging.DEBUG, traceback_)


            self.add_to_logging(
                logging.DEBUG,
                "[{0}] Response body: {1}".format(self.request_id, json_secure(json_res_body)),
            )

            res_body = json.dumps(json_res_body).encode("utf-8")
            return self.make_response(response, res_body)

        return response

    @staticmethod
    def make_response(response: Response, body: bytes) -> Response:
        """Create response method"""
        headers = MutableHeaders(scope={"headers": response.headers})
        headers.raw["content-length"] = str(len(body))

        return Response(
            content=body,
            status_code=response.status_code,
            headers=headers.raw,
            media_type=response.media_type,
        )

    @staticmethod
    def is_response_body_json(response: Response) -> bool:
        """Checked response content type is json"""
        return response.headers["content-type"] == "application/json"


class ExceptionMiddleware(BaseHTTPMiddleware):

    async def dispatch(
        self, request: Request, call_next: RequestResponseEndpoint
    ) -> Union[Response, JSONResponse]:
        try:
            response = await call_next(request)
        except NotImplementedError as e:
            status_code = HTTP_501_NOT_IMPLEMENTED
            response = self.make_response(e, status_code, logging.INFO)
        except AuthenticationError as e:
            status_code = HTTP_401_UNAUTHORIZED
            response = self.make_response(e, status_code, logging.INFO)
        except Exception as e:
            status_code = HTTP_500_INTERNAL_SERVER_ERROR
            response = self.make_response(e, status_code)

        return response

    def make_response(self, exc: BaseSSAException, status_code: int, exc_level=logging.ERROR) -> Response:
        """Create response method"""

        return JSONResponse(
            status_code=status_code,
            content={
                "detail": str(exc),
                "type": self.get_full_exc_path(exc),
                "traceback": traceback.format_exc(),
                "exc_level": exc_level,
            },
        )

    @staticmethod
    def get_full_exc_path(exc: BaseSSAException) -> str:
        """Returns the full path of exception class from an exception object"""

        exception_class = exc.__class__
        module_name = exception_class.__module__
        class_name = exception_class.__name__
        full_path = f"{module_name}.{class_name}"

        return full_path
