import configparser
import logging
import os

from scripts.ssl import generate_selfsigned_cert

# FIXME: данная реализация предназначена для корректного выполнения тестов
try:
    from environment.checks import PlatformEnvironment
except ImportError:
    from session_agent.environment.checks import PlatformEnvironment

logger = logging.getLogger("config.handler")

if PlatformEnvironment.is_linux():
    from constants.linux.constants import ConstLinuxEnv as const
if PlatformEnvironment.is_windows():
    from constants.windows.constants import ConstWindowsEnv as const


def config_exist() -> bool:
    """Checking if config is existing"""
    return os.path.exists(const.CONFIG_FILE_PATH)


def _cert_path():
    """Getting SSL certificate path from config"""
    return ConfigHandler.select("SSL", "CERTIFICATE_PATH")


def _private_key_path():
    """Getting SSL private key path from config"""
    return ConfigHandler.select("SSL", "PRIVATE_KEY_PATH")


def _cert_exist():
    """Checking if SSL certificate exist"""
    return os.path.exists(_cert_path())


def _private_key_exist():
    """Checking if SSL private key exist"""
    return os.path.exists(_private_key_path())


def using_default_cert() -> bool:
    """Checking if SSL certificate and private key exists and located in default paths.

    This verification is required to verify that
    all conditions are met before generating a
    self-signed certificate.

    """
    cert_is_default = _cert_path() == const.SESSION_AGENT_CONFIG["SSL"]["CERTIFICATE_PATH"]
    key_is_default = _private_key_path() == const.SESSION_AGENT_CONFIG["SSL"]["PRIVATE_KEY_PATH"]

    if cert_is_default and key_is_default:
        cert_key_not_exists = all([not _cert_exist(), not _private_key_exist()])

        return cert_key_not_exists

    return False


def _check_cert_files() -> None:
    """Checking if cert files is existing"""
    if using_default_cert():
        generate_selfsigned_cert(key_file=_private_key_path(), cert_file=_cert_path())

    if not _cert_exist():
        err_msg = "Файл сертификата не найден. " \
                  f"Поместите файл в директорию: {_cert_path()}. " \
                  f"Или укажите путь до сертификата в конфигурационном файле: {const.CONFIG_FILE_PATH}."
        logger.error(err_msg)
        raise FileNotFoundError(err_msg)

    if not _private_key_exist():
        err_msg = "Файл приватного ключа не найден. " \
                  f"Поместите файл в директорию: {_private_key_path()}. " \
                  f"Или укажите путь до приватного ключа в конфигурационном файле: {const.CONFIG_FILE_PATH}."
        logger.error(err_msg)
        raise FileNotFoundError(err_msg)


class ConfigHandler:
    """Session agent config handler.
    Arguments with default values in methods are required for tests.
    """
    @staticmethod
    def _log_setup() -> None:
        """Log warning and config params."""
        with open(const.CONFIG_FILE_PATH, "r", encoding="utf-8") as f:
            to_log = "Session agent configuration file wasn't found. " \
                     "Creating session_agent.ini with following params:\n" \
                     f"Path: {const.CONFIG_FILE_PATH}\n" \
                     "Parameters:\n"

            logger.warning(to_log + f.read().rstrip())

    @staticmethod
    def _actualize(dict_config: dict, to_log: bool) -> None:
        """Adds new sections and options to already existing outdated config"""
        ssa_conf = configparser.ConfigParser()
        # Without this option values in config will be lowercase.
        ssa_conf.optionxform = str
        ssa_conf.read(const.CONFIG_FILE_PATH)

        for section, options in dict_config.items():
            # If section doesn't exist: add it.
            if not ssa_conf.has_section(section):
                ssa_conf.add_section(section)
                if to_log:
                    logger.warning(f"New section in configuration file: {section}.")
            # If option in section doesn't exist: add it.
            for opt_key, opt_value in options.items():
                if not ssa_conf.has_option(section, opt_key):
                    ssa_conf.set(section, opt_key, str(opt_value))
                    if to_log:
                        logger.warning(
                            f"New option in configuration file: {opt_key}. Section: {section}."
                        )

        with open(const.CONFIG_FILE_PATH, "w", encoding="utf-8") as conf:
            ssa_conf.write(conf)

    @staticmethod
    def setup(dict_config=const.SESSION_AGENT_CONFIG, to_log=True) -> None:
        """Creating session_agent.ini: configuration file based on SESSION_AGENT_CONFIG"""
        if config_exist():
            ConfigHandler._actualize(dict_config, to_log)
        else:
            with open(const.CONFIG_FILE_PATH, "w", encoding="utf-8") as conf:
                ssa_conf = configparser.ConfigParser()
                # Without this option values in config will be lowercase.
                ssa_conf.optionxform = str
                for section, options in dict_config.items():
                    ssa_conf.add_section(section)
                    for opt_key, opt_value in options.items():
                        ssa_conf.set(section, opt_key, str(opt_value))
                ssa_conf.write(conf)
            # Log warning and config params if session_agent.ini doesn't exist.
            if to_log:
                ConfigHandler._log_setup()
        # Required for auth work
        _check_cert_files()

    @staticmethod
    def select(section: str, option: str, dict_config=const.SESSION_AGENT_CONFIG, to_log=True) -> (str, bool):
        """Select option from config file"""
        ssa_conf = configparser.RawConfigParser()
        ssa_conf.read(const.CONFIG_FILE_PATH)
        # Checking option type. Selecting different types requires specific methods.
        try:
            default_option = dict_config[section][option]
            if isinstance(default_option, bool):
                ssa_param = ssa_conf.getboolean(section, option)
            elif isinstance(default_option, int):
                ssa_param = ssa_conf.getint(section, option)
            elif isinstance(default_option, float):
                ssa_param = ssa_conf.getfloat(section, option)
            else:
                ssa_param = ssa_conf.get(section, option)

        except ValueError as ex:
            # If the option type in real config not equivalent to the option type in default config:
            # log error and raise the exception.
            option_type = type(default_option)
            if to_log:
                logger.error("Некорректный тип данных в конфигурационном файле.")
                logger.error(f"Параметр {section} - {option} должен быть следующего типа: {option_type}.")
            raise ex

        except (KeyError, configparser.NoSectionError, configparser.NoOptionError) as ex:
            if to_log:
                logging.error(f"В конфигурационном файле отсутствует параметр {section} - {option}, "
                              "необходимый для корректной работы сессионного агента.")
            raise ex

        return ssa_param
