from container import Container
from dependency_injector.wiring import Provide, inject
from excepts import AuthenticationError
from fastapi import Depends
from fastapi import Request
from fastapi_utils.inferring_router import InferringRouter
from utils.secure import JWTAuthorization

from models.base import (
    HTTP_500_RESPONSE_MODEL,
    HTTP_401_RESPONSE_MODEL,
)

from models.auth import (
    TokenRefresh,
    AuthUserModel,
    AuthResponse,
)


auth_router = InferringRouter()


@auth_router.post("/auth", response_model=AuthResponse,
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL, })
@inject
def login(request: Request, user: AuthUserModel,
          service=Depends(Provide[Container.service]),
          auth: JWTAuthorization = Depends(Provide[Container.jwt_auth])
          ):
    """method to handle connection request to rds.

    Args:
        request (Request): http request
        user (AuthUserModel): auth params
        service (StalService|RDSService): service layer

    Returns:
        response (dict): json response
    """

    try:
        is_connected = service.test_connect(user)

        if not is_connected:
            raise AuthenticationError("Учетные данные не прошли проверку")

        login, domain = user.login, user.domain
        ip_addr = request.client.host
        payload = {'login': login, 'domain': domain, 'ip_addr': ip_addr}
        response = auth.create_tokens(payload)

    except Exception as ex:
        raise AuthenticationError(ex)

    return response


@auth_router.post("/auth/refresh", response_model=AuthResponse,
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL, })
@inject
def refresh(params: TokenRefresh,
            auth: JWTAuthorization = Depends(Provide[Container.jwt_auth])
            ):
    """method for handling a request to refresh JWT-token

    Returns:
        response (JSONResponse): json response
    """
    refresh_token = params.refresh_token

    response = auth.refresh(refresh_token)

    return response
