from fastapi import status, BackgroundTasks
from fastapi_utils.cbv import cbv
from fastapi.responses import JSONResponse
from fastapi_utils.inferring_router import InferringRouter

from models.base import (
    BaseDelegateModel,
    HealthCheckResponse,
    VersionCheckResponse,
)

from threading import Thread
from services.agent import AgentService
import _version
import logging
from env_config import SSA_URL_DELEGATE_NAME, SSA_URL_DELEGATE
logger = logging.getLogger(__name__)

# Const for static app status response
RESPONSE_OK = {"status": "ok"}

PATH_DELEGATE = "/agent/delegate"
base_router = InferringRouter()


@cbv(base_router)
class BaseHandler:

    @base_router.get("/public/status", response_model=HealthCheckResponse)
    def healthcheck(self):
        """function to check the health of the server

        Returns:
            response (JSONResponse): json response
        """

        return RESPONSE_OK

    @base_router.get("/version", response_model=VersionCheckResponse)
    def version_check(self):
        """method for handling a request to get version

        Returns:
            response (JSONResponse): json response
        """
        version = _version.__version__
        response = {"version": version}

        return response

    @base_router.post(PATH_DELEGATE)
    def delegate(self, params: BaseDelegateModel, background_tasks: BackgroundTasks) -> dict:
        """method for manage agent work delegate

        Returns:
            response (JSONResponse): json response
        """

        request_body = params.json()
        log_body = "VM Agent has been notified of the delegation " \
                   f"acceptance from SSA. HTTP 200 POST '{PATH_DELEGATE}' " \
                   f"{SSA_URL_DELEGATE_NAME}:'{SSA_URL_DELEGATE}' " \
                   f"request body:'{request_body}'"

        # accept work delegate from agent
        process = AgentService(params.url).process
        background_process = Thread(target=process)
        background_process.start()

        background_tasks.add_task(logger.info, log_body)

        return JSONResponse(status_code=status.HTTP_200_OK,
                            content=RESPONSE_OK)
