from fastapi import Depends

from fastapi_utils.inferring_router import InferringRouter

from fastapi import Request, Depends
from excepts import OSAPIError, AuthenticationError

from models.base import (
    HTTP_500_RESPONSE_MODEL,
    HTTP_501_RESPONSE_MODEL,
    HTTP_401_RESPONSE_MODEL,
)

from models.meta import (
    MetaApps,
    MetaAvailableApps,
    MetaLogin,
    MetaLogout,
    MetaLogoutModel,
    MetaProgramRequestModel,
    MetaSessionRequest,
    MetaSessions,
    MetaTest,
    MetaTestConnect,
    MetaUserModel,
    MetaInstallRdsRole,
    MetaPublishAppModel,
    MetaDeleteApp,
    MetaPublishApp,
)
from services.meta import MetaService
from container import Container
from dependency_injector.wiring import Provide, inject
from core.secure import check_auth

meta_router = InferringRouter()

"""
Class for handling requests for Meta

Handled errors:
- OSAPIError (из модуля excepts): the error that occurs
  when executing a request API ОС
- AuthenticationError (из модуля excepts): the error that occurs
  during authentication
"""


@meta_router.get("/test", response_model=MetaTest)
@inject
def check_instance(service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Checking the health of the meta server
    """
    service_name = service.name
    response = {"service_name": service_name}

    return response


@meta_router.get("/sessions", dependencies=[Depends(check_auth)], response_model=MetaSessions,
                 responses={**HTTP_500_RESPONSE_MODEL,
                            **HTTP_401_RESPONSE_MODEL,})
@inject
def sessions(service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Getting sessions list
    """
    try:
        sessions = service.get_sessions()
        response = {"sessions": sessions}
    except Exception as ex:
        raise OSAPIError(ex)

    return response


@meta_router.post("/login", response_model=MetaLogin,
                  responses={**HTTP_401_RESPONSE_MODEL,})
@inject
def login(
        params: MetaUserModel,
        service: MetaService = Depends(Provide[Container.meta_service])
):
    """
    User authorization

    - **login**: user login
    - **domain**: user domain
    - **password**: user password
    """
    try:
        is_authenticated = service.login_user(params)

        response = {"auth_done": is_authenticated}
    except Exception as ex:
        raise AuthenticationError(ex)

    return response


@meta_router.post("/testconnect", response_model=MetaTestConnect,
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL,})
@inject
def test_connect(
        params: MetaUserModel,
        service: MetaService = Depends(Provide[Container.meta_service])
):
    """
    Checking connect user

    - **login**: user login
    - **domain**: user domain
    - **password**: user password
    """
    try:
        is_connected = service.test_connect(params)
        response = {"is_connected": is_connected}
    except Exception as ex:
        raise AuthenticationError(ex)

    return response


@meta_router.post("/logout", dependencies=[Depends(check_auth)], response_model=MetaLogout,
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL,})
@inject
def logout_user(
        params: MetaLogoutModel,
        service: MetaService = Depends(Provide[Container.meta_service])
):
    """
    Disconnecting user

    - **session_id**: session id for logout
    """
    try:
        service.logout_user(params)

        response = {"disconnect_done": True}
    except Exception as ex:
        raise OSAPIError(ex)

    return response


@meta_router.get("/published_apps", dependencies=[Depends(check_auth)], response_model=MetaApps,
                 responses={**HTTP_500_RESPONSE_MODEL,
                            **HTTP_401_RESPONSE_MODEL,})
@inject
def published_apps(service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Getting published applications
    """
    try:
        apps = service.get_remote_apps()

        response = {"apps": apps}
    except Exception as ex:
        raise OSAPIError(ex)

    return response


@meta_router.post("/applications", response_model=MetaPublishApp, dependencies=[Depends(check_auth)],
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_501_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL,})
@inject
def publish_app(params: MetaPublishAppModel, service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Adding an application on Windows Server to the list of available

    - **name**: application name
    - **path**: path to application
    """
    try:
        service.publish_app(params)

        response = {"is_created": True}
    except Exception as ex:
        raise OSAPIError(ex)

    return response

@meta_router.delete("/applications/{name}", response_model=MetaDeleteApp,
                    dependencies=[Depends(check_auth)],
                    responses={**HTTP_500_RESPONSE_MODEL,
                               **HTTP_501_RESPONSE_MODEL,
                               **HTTP_401_RESPONSE_MODEL,})
@inject
def delete_app(name: str, service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Removing an application on Windows Server from the list of available

    - **name**: application name
    """
    try:
        service.delete_app(name)

        response = {"is_deleted": True}
    except Exception as ex:
        raise OSAPIError(ex)

    return response

@meta_router.get("/available_apps", dependencies=[Depends(check_auth)], response_model=MetaAvailableApps,
                 responses={**HTTP_500_RESPONSE_MODEL,
                            **HTTP_501_RESPONSE_MODEL,
                            **HTTP_401_RESPONSE_MODEL,})
@inject
def check_available_apps(service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Getting publishable Windows Server applications
    """
    try:
        app_list = service.get_list_startapps()
        response = {"available_apps": app_list}
    except Exception as ex:
        raise OSAPIError(ex)

    return response


@meta_router.post("/sessionrequest", dependencies=[Depends(check_auth)], response_model=MetaSessionRequest,
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_501_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL,})
@inject
def session_request(user: MetaUserModel, service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Method for creating user session in STAL
    """
    # session_id может быть -1, если сессия не была создана
    session_id = service.session_request(user)  # dbus.Int32

    if session_id == -1:
        raise OSAPIError("STAL session not created!")

    result = {"session_id": session_id}

    return result


@meta_router.post("/programrequest", dependencies=[Depends(check_auth)], response_model=MetaSessionRequest,
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_501_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL,})
@inject
def program_request(data: MetaProgramRequestModel, service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Method for creating user session in STAL in single program mode
    """
    session_id = service.program_request(data)  # dbus.Int32

    if session_id == -1:
        raise OSAPIError("STAL session not created!")

    result = {"session_id": session_id}

    return result


@meta_router.post("/install_rds_role", dependencies=[Depends(check_auth)], response_model=MetaInstallRdsRole,
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_501_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL,})
@inject
def install_rds_role(service: MetaService = Depends(Provide[Container.meta_service])):
    """
    Method for handling a request to install rds role on powershell
    """

    response = service.install_rds_role()

    return response
