from fastapi import Depends
from fastapi_utils.inferring_router import InferringRouter

from excepts import AuthenticationError, OSAPIError
from models.base import (
    HTTP_500_RESPONSE_MODEL,
    HTTP_401_RESPONSE_MODEL,
)
from pywintypes import error as pywin_error
from wmi import x_wmi

from models.rds import (
    RdsLogoutModel,
    RdsRemoteAppModel,
    RdsUserModel,
    RemoteAppUsersModel,
    CheckInstanceResponse,
    SessionsResponse,
    LoginResponse,
    LogoutUserResponse,
    TestConnectResponse,
    CollectionsResponse,
    RdsRemoteAppResponse,
    RemoteAppUsersResponse,
    CollectionInfoResponse
)

from utils.rds import check_instance_is_rds
from services.rds import RdsService
from container import Container
from dependency_injector.wiring import Provide, inject
from core.secure import check_auth

rds_router = InferringRouter()


@rds_router.get("/rds/test", response_model=CheckInstanceResponse,
                responses={**HTTP_500_RESPONSE_MODEL, })
def check_rds_instance():
    """
    Method for handling server rds check request.

    Returns:
        response (dict): json response
    """
    try:
        result = check_instance_is_rds()
    except Exception as ex:
        raise OSAPIError(ex)

    response = {"is_rds": result}
    return response


@rds_router.get("/rds/sessions",
                response_model=SessionsResponse,
                dependencies=[Depends(check_auth)],
                responses={**HTTP_500_RESPONSE_MODEL,
                           **HTTP_401_RESPONSE_MODEL, })
@inject
def rds_sessions(service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling a request to get rds user sessions.
    """
    try:
        result = service.get_sessions()
    except x_wmi as ex:
        raise OSAPIError(ex)

    response = {"sessions": result}
    return response


@rds_router.post("/rds/login", response_model=LoginResponse,
                 responses={**HTTP_401_RESPONSE_MODEL, })
@inject
def rds_login(params: RdsUserModel, service: RdsService = Depends(Provide[Container.service])):
    """
    Method to handle connection request to rds.
    """
    try:
        result = service.login_user(params)
    except pywin_error as ex:
        raise AuthenticationError(ex)

    response = {"auth_done": result}
    return response


@rds_router.post("/rds/testconnect", response_model=TestConnectResponse,
                 responses={**HTTP_401_RESPONSE_MODEL, })
@inject
def rds_test_connect(params: RdsUserModel, service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling request for a test connection to RDS.
    """
    try:
        result = service.test_connect(params)
    except pywin_error as ex:
        raise AuthenticationError(ex)

    response = {"is_connected": result}
    return response


@rds_router.post("/rds/disconnect",
                 response_model=LogoutUserResponse,
                 dependencies=[Depends(check_auth)],
                 responses={**HTTP_500_RESPONSE_MODEL,
                            **HTTP_401_RESPONSE_MODEL, })
@inject
def rds_logout_user(params: RdsLogoutModel, service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling a request to disconnect a user.
    """
    try:
        result = service.logout_user(params)
    except Exception as ex:
        raise OSAPIError(ex)

    response = {"disconnect_done": result}
    return response


@rds_router.get("/rds/collections",
                response_model=CollectionsResponse,
                dependencies=[Depends(check_auth)],
                responses={**HTTP_500_RESPONSE_MODEL,
                           **HTTP_401_RESPONSE_MODEL, })
@inject
def rds_collections(service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling a request to get rds collections.
    """
    try:
        result = service.get_collections()
    except x_wmi as ex:
        raise OSAPIError(ex)

    response = {"collections": result}
    return response


@rds_router.post("/rds/apps",
                 response_model=RdsRemoteAppResponse,
                 dependencies=[Depends(check_auth)],
                 responses={**HTTP_500_RESPONSE_MODEL,
                            **HTTP_401_RESPONSE_MODEL, })
@inject
def rds_apps(params: RdsRemoteAppModel, service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling a request to get published remote apps.
    """
    try:
        result = service.get_remote_apps(params)
    except x_wmi as ex:
        raise OSAPIError(ex)

    response = {"apps": result}
    return response


@rds_router.post("/rds/apps/users",
                 response_model=RemoteAppUsersResponse,
                 dependencies=[Depends(check_auth)],
                 responses={**HTTP_500_RESPONSE_MODEL,
                            **HTTP_401_RESPONSE_MODEL, })
@inject
def rds_app_users(params: RemoteAppUsersModel, service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling a request to getting users that have access
    to published app by collection name and remote app alias.
    """
    try:
        result = service.get_remote_app_users(params)
    except Exception as ex:
        raise OSAPIError(ex)

    response = {"app_users": result}
    return response


@rds_router.get("/rds/collection_info/{collection_name}",
                response_model=CollectionInfoResponse,
                dependencies=[Depends(check_auth)],
                responses={**HTTP_500_RESPONSE_MODEL,
                           **HTTP_401_RESPONSE_MODEL, })
@inject
def collection_info(collection_name: str, service: RdsService = Depends(Provide[Container.service])):
    """
    Method for handling a request to get collection info by collection name.
    """
    try:
        result = service.get_collection_info(collection_name)
    except x_wmi as ex:
        raise OSAPIError(ex)

    response = {"collection_info": result}
    return response
