from fastapi import Depends
from fastapi_utils.inferring_router import InferringRouter

import dbus
from excepts import AuthenticationError, OSAPIError
from models.base import (
    HTTP_500_RESPONSE_MODEL,
    HTTP_401_RESPONSE_MODEL,
)
from models.stal import (
    StalLogoutModel,
    StalUserModel,
    StalRemoteAppModel,
    VersionResponse,
    StatusResponse,
    SessionsResponse,
    LoginUserResponse,
    DisconnectUserResponse,
    ShutdownUserResponse,
    TestConnectResponse,
    SessionRequestResponse,
    ProgramRequestResponse,
    AppsResponse,
)
from services.stal import StalService
from container import Container
from dependency_injector.wiring import Provide, inject
from core.secure import check_auth

stal_router = InferringRouter()


@stal_router.get("/stal/versions", response_model=VersionResponse)
@inject
def version(service: StalService = Depends(Provide[Container.service])):
    """
    Method for gets STAL version.
    """
    try:
        result = service.version()
    except Exception as ex:
        raise OSAPIError(ex)

    response = {"version": result, "message": "OK!"}
    return response


@stal_router.get("/rds/test",  # FIXME: "/stal/test"
                 response_model=StatusResponse,
                 responses={**HTTP_500_RESPONSE_MODEL, })
@inject
def status(service: StalService = Depends(Provide[Container.service])):
    """
    Method for gets STAL status.
    """
    try:
        result = service.status()
    except Exception as ex:
        raise OSAPIError(ex)

    response = {"is_stal": result}
    return response


@stal_router.get("/rds/sessions",  # FIXME: "/stal/sessions"
                 response_model=SessionsResponse,
                 dependencies=[Depends(check_auth)],
                 responses={**HTTP_500_RESPONSE_MODEL,
                            **HTTP_401_RESPONSE_MODEL, })
@inject
def sessions(service: StalService = Depends(Provide[Container.service])):
    """
    Method for gets a list of running sessions.
    """
    try:
        result = service.get_sessions()
    except Exception as ex:
        raise OSAPIError(ex)

    response = {"sessions": result}
    return response


@stal_router.post("/rds/login",  # FIXME: "/stal/login"
                  response_model=LoginUserResponse,
                  responses={**HTTP_401_RESPONSE_MODEL, })
@inject
def login_user(
        user: StalUserModel,
        service: StalService = Depends(Provide[Container.service])
):
    """
    Method for authentication user
    """
    try:
        result = service.login_user(user)
    except Exception as ex:
        raise AuthenticationError(ex)

    response = {"auth_done": result}
    return response


@stal_router.post("/stal/disconnect_",
                  response_model=DisconnectUserResponse,
                  dependencies=[Depends(check_auth)],
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL, })
@inject
def disconnect_user(
        user: StalLogoutModel,
        service: StalService = Depends(Provide[Container.service])
):
    """
    Method for disconnect user
    """
    try:
        result = service.disconnect_user(user)
    except Exception as ex:
        raise OSAPIError(ex)

    response = {"message": "OK!" if result else "BAD!"}
    return response


@stal_router.post("/rds/disconnect",  # FIXME: "/stal/shutdown"
                  response_model=ShutdownUserResponse,
                  dependencies=[Depends(check_auth)],
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL, })
@inject
def shutdown_user(
        user: StalLogoutModel,
        service: StalService = Depends(Provide[Container.service])
):
    """
    Method for shutdown user session
    """
    try:
        result = service.logout_user(user)
    except Exception as ex:
        raise OSAPIError(ex)

    response = {"disconnect_done": result}
    return response


@stal_router.post("/rds/testconnect",  # FIXME: "/stal/testconnect"
                  response_model=TestConnectResponse,
                  responses={**HTTP_500_RESPONSE_MODEL, })
@inject
def test_connect(
        user: StalUserModel,
        service: StalService = Depends(Provide[Container.service])
):
    """
    Method for check user in STAL
    """
    try:
        result = service.test_connect(user)
    except Exception as ex:
        raise OSAPIError(ex)

    response = {"is_connected": result}
    return response


@stal_router.post("/stal/sessionrequest",
                  response_model=SessionRequestResponse,
                  dependencies=[Depends(check_auth)],
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL, })
@inject
def session_request(
        user: StalUserModel,
        service: StalService = Depends(Provide[Container.service])
):
    """
    Method for creating user session in STAL
    """
    try:
        result = service.session_request(user)
    except dbus.exceptions.DBusException as ex:
        raise OSAPIError(ex)

    response = {"session_id": result}
    return response


@stal_router.post("/stal/programrequest",
                  response_model=ProgramRequestResponse,
                  dependencies=[Depends(check_auth)],
                  responses={**HTTP_500_RESPONSE_MODEL,
                             **HTTP_401_RESPONSE_MODEL, })
@inject
def program_request(
        data: StalRemoteAppModel,
        service: StalService = Depends(Provide[Container.service])
):
    """
    Method for creating user session in STAL in single program mode
    """
    try:
        result = service.program_request(data)
    except dbus.exceptions.DBusException as ex:
        raise OSAPIError(ex)

    response = {"session_id": result}
    return response


@stal_router.get("/stal/apps",
                 response_model=AppsResponse,
                 dependencies=[Depends(check_auth)],
                 responses={**HTTP_500_RESPONSE_MODEL,
                            **HTTP_401_RESPONSE_MODEL, })
@inject
def stal_apps(service: StalService = Depends(Provide[Container.service])):
    """
    Method for to get published remote apps.
    """
    try:
        result = service.get_remote_apps()
    except Exception as ex:
        raise OSAPIError(ex)

    response = {"apps": result}
    return response
