from logging.handlers import RotatingFileHandler
import os
import tarfile

from constants.common_const import ConstConfig as const
from default import config_exist, ConfigHandler as config
from environment.checks import PlatformEnvironment

# Т.к. логи инициализируются до создания конфигурационного файла, при первом запуске берутся
# значения по умолчанию. При последующих: значения из конфигурационного файла.
if config_exist():
    LOG_LEVEL = config.select("LOGGING", "LEVEL")
    LOG_BACKUP_COUNT = config.select("LOGGING", "ROTATION_LIMIT")
    LOG_MAX_BYTES = config.select("LOGGING", "MAX_FILE_SIZE")
else:
    LOG_LEVEL = const.SESSION_AGENT_CONFIG["LOGGING"]["LEVEL"]
    LOG_BACKUP_COUNT = const.SESSION_AGENT_CONFIG["LOGGING"]["ROTATION_LIMIT"]
    LOG_MAX_BYTES = const.SESSION_AGENT_CONFIG["LOGGING"]["MAX_FILE_SIZE"]

if PlatformEnvironment.is_linux():
    LOG_FILE_PATH = "/var/log/termidesk/session-agent.log"

if PlatformEnvironment.is_windows():
    PROGRAM_DATA = os.path.expandvars(r"%PROGRAMDATA%")
    LOG_FILE_PATH = os.path.join(
        PROGRAM_DATA, "UVEON", "Termidesk Session Agent", "log", "session-agent.log"
    )


LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "default": {
            "()": "uvicorn.logging.DefaultFormatter",
            "fmt": "%(levelprefix)s %(message)s",
            "use_colors": None,
        },
        "access": {
            "()": "uvicorn.logging.AccessFormatter",
            "fmt": '%(levelprefix)s %(client_addr)s - "%(request_line)s" %(status_code)s',
        },
        "file": {
            "format": "%(asctime)s.%(msecs)03d %(levelname)-6s %(name)s - %(message)s",
        },
    },
    "handlers": {
        "default": {
            "formatter": "default",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stderr",
        },
        "access": {
            "formatter": "access",
            "class": "logging.StreamHandler",
            "stream": "ext://sys.stdout",
        },
        "file": {
            "formatter": "file",
            "class": "logging_config.CompressingFileHandler",
            "filename": LOG_FILE_PATH,
            "maxBytes": LOG_MAX_BYTES,
            "backupCount": LOG_BACKUP_COUNT,
        },
    },
    "loggers": {
        "uvicorn": {
            "handlers": ["default"],
            "level": LOG_LEVEL,
            "propagate": False,
        },
        "uvicorn.error": {"level": LOG_LEVEL},
        "uvicorn.access": {
            "handlers": ["access"],
            "level": LOG_LEVEL,
            "propagate": False,
        },
    },
    "root": {
        "handlers": ["default", "file"],
        "level": LOG_LEVEL,
        "propagate": False,
    },
}


class CompressingFileHandler(RotatingFileHandler):
    """Improved RotatingFileHandler with log compression mechanism"""
    def _get_archive(self) -> str:
        """Get unique archive path"""
        i = 1
        while True:
            archive_path = self.rotation_filename("%ss.old.%d.tar.gz" % (self.baseFilename, i))
            if os.path.exists(archive_path):
                i += 1
            else:
                break
        return archive_path

    def _ready_to_compress(self) -> bool:
        """Checking if all logs exist before compression"""
        try:
            all_logs_exist = []
            for i in range(1, self.backupCount + 1):
                if os.path.exists(self.rotation_filename("%s.%d" % (self.baseFilename, i))):
                    all_logs_exist.append(True)
                else:
                    all_logs_exist.append(False)
                    break

            return all(all_logs_exist)
        except:
            return False

    def _compress_logs(self):
        """Compress logs with tar.gz"""
        with tarfile.open(self._get_archive(), "w:gz") as tar:
            for i in range(1, self.backupCount + 1):
                log_file = self.rotation_filename("%s.%d" % (self.baseFilename, i))
                file_name = os.path.basename(log_file)
                tar.add(log_file, arcname=file_name)
                os.remove(log_file)

    def doRollover(self):
        """
        Do a rollover, as described in __init__().
        """
        if self.stream:
            self.stream.close()
            self.stream = None
        if self.backupCount > 0:

            if self._ready_to_compress():
                self._compress_logs()

            for i in range(self.backupCount - 1, 0, -1):
                sfn = self.rotation_filename("%s.%d" % (self.baseFilename, i))
                dfn = self.rotation_filename("%s.%d" % (self.baseFilename,
                                                        i + 1))
                if os.path.exists(sfn):
                    if os.path.exists(dfn):
                        os.remove(dfn)
                    os.rename(sfn, dfn)

            dfn = self.rotation_filename(self.baseFilename + ".1")
            if os.path.exists(dfn):
                os.remove(dfn)

            self.rotate(self.baseFilename, dfn)
        if not self.delay:
            self.stream = self._open()
