import json
from typing import Any, Dict, List

import dbus

from constants.linux.constants import ConstStal
from models.meta import MetaLogoutModel, MetaProgramRequestModel, MetaPublishAppModel, MetaUserModel
from utils.base import extract_ip


class MetaService:
    """
    Сервисный слой для взаимодействия со STAL
    """

    __bus = None
    __stal = None
    __stal_iface = None
    __session_info: dict = {}
    __name = "stal"

    # Class implemented via Singleton
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super(MetaService, cls).__new__(cls)
        return cls._instance
    
    # DI для проброса констант
    def __init__(self, const: ConstStal):
        self.DEPTH = const.DEPTH
        self.DISPLAY_WIDTH = const.DISPLAY_WIDTH
        self.DISPLAY_HEIGHT = const.DISPLAY_HEIGHT
        self.SESSION_STATES = const.SESSION_STATES

    @property
    def stal_iface(self) -> dbus.Interface:
        if not self.__stal_iface:
            self.__bus = dbus.SystemBus()
            self.__stal = self.__bus.get_object("ru.uveon.stal", "/ru/uveon/stal")
            self.__stal_iface = dbus.Interface(self.__stal, "ru.uveon.stal")
            return self.__stal_iface

        return self.__stal_iface

    @property
    def name(self) -> str:
        return self.__name

    def get_session_info(self, session_id: dbus.UInt32) -> dbus.String:
        """
        Gets session information
        """
        session_info: dbus.String = self.stal_iface.stalSessionInfoJson(session_id)
        return session_info

    def __load_session_info(self, session_id: dbus.UInt32) -> None:
        """
        Records session information
        """
        info: dbus.String = self.get_session_info(session_id)
        self.__session_info = json.loads(info)

    def __clear_session_info(self) -> None:
        """
        Removes session information
        """
        self.__session_info = None

    def __get_session_state(self) -> str:
        """
        Get session state
        """
        is_session_active = self.__session_info.get('remote:connected')
        if is_session_active is None:
            # FIXME:Для stal меньше v.1.1 проверяем наличие ключа 'pid3'
            is_session_active = self.__session_info["pid3"] != 0
        return self.SESSION_STATES[is_session_active]

    def __get_session_is_app(self) -> bool:
        """
        Get the value of whether the session is an application
        Parameter value 'seamless:mode':
        if mode == false - terminal
        if mode == true - app
        """
        is_app_mode = self.__session_info["seamless:mode"]
        if isinstance(is_app_mode, str):  # special for STAL V0.3
            is_app_mode = is_app_mode == "true"
        return is_app_mode

    def get_sessions(self) -> List[Dict[str, Any]]:
        """
        Method for gets STAL sessions
        """
        # Example:
        # dbus.Array(
        #     [
        #         dbus.Struct(
        #             (dbus.String("test1"), dbus.UInt32(59101)), signature=None
        #         )
        #     ],
        #     signature=dbus.Signature("(su)"),
        # )
        dbus_sessions: dbus.Array[dbus.Struct] = self.stal_iface.stalSessions()
        host_server = extract_ip()
        sessions = []
        for session in dbus_sessions:
            username, session_id = session
            self.__load_session_info(session_id)
            state = self.__get_session_state()
            is_desktop = not self.__get_session_is_app()
            self.__clear_session_info()

            sessions.append(
                {
                    "username": username,
                    "session_id": session_id,
                    "host_server": host_server,  # for logoff in future
                    "state": state,
                    "is_desktop": is_desktop,
                }
            )

        return sessions

    def test_connect(self, params: MetaUserModel) -> dbus.Boolean:
        """
        Method for authentication
        """
        is_success: dbus.Boolean = self.stal_iface.stalTestConnect(
            params.login,  # user
            params.domain,  # domain
            params.password,  # pass
        )
        return is_success

    def login_user(self, params: MetaUserModel) -> dbus.Boolean:
        """
        Method for start of a new user session
        """
        is_success: dbus.Boolean = self.stal_iface.stalLogin(
            params.login,  # user
            params.domain,  # domain
            params.password,  # pass
        )
        return is_success

    def logout_user(self, params: MetaLogoutModel) -> dbus.Boolean:
        """
        Method for shutdown user session
        Session closes assync ~600ms
        """
        is_success: dbus.Boolean = self.stal_iface.stalSessionShutdown(
            params.session_id,
        )
        return is_success

    def session_request(self, params: MetaUserModel) -> dbus.Int32:
        """
        Method for creating user session in STAL
        """
        # session_id может быть -1, если сессия не была создана
        session_id: dbus.Int32 = self.stal_iface.busSessionRequest(
            params.login,  # user
            params.password,  # pass
            self.DEPTH,  # depth
            self.DISPLAY_WIDTH,  # width
            self.DISPLAY_HEIGHT,  # height
        )
        return session_id

    def program_request(self, params: MetaProgramRequestModel) -> dbus.Int32:
        """
        Method for creating user session in STAL in single program mode
        """
        session_id: dbus.Int32 = self.stal_iface.busProgramRequest(
            params.login,  # user
            params.domain,  # domain
            params.password,  # pass
            self.DISPLAY_WIDTH,  # width
            self.DISPLAY_HEIGHT,  # height
            params.program,  # program
        )
        return session_id

    def get_remote_apps(self) -> dbus.Array:
        """Getting published remote apps from STAL.

        Returns:
            result (dict): json response
        """
        remote_apps: dbus.Array = self.stal_iface.stalProgramList()
        return remote_apps

    def publish_app(self, params: MetaPublishAppModel) -> None:
        """Application Publishing method on STAL

        Args:
            params (MetaPublishAppModel): publish params
        """
        raise NotImplementedError("The method is not intended to be called on this OS.")

    def delete_app(self, name: str) -> None:
        """Deleting an application to STAL

        Args:
            name (str): application name
        """
        raise NotImplementedError("The method is not intended to be called on this OS.")

    def get_list_startapps(self) -> Dict[str, str]:
        """Getting list of apps from Start Menu. Available apps for publish"""
        raise NotImplementedError("The method is not intended to be called on this OS.")

    def install_rds_role(self) -> dict:
        """Configuring the RDS Role

        Returns:
            result (dict): json response
        """
        raise NotImplementedError("The method is not intended to be called on this OS.")
