﻿import os
import eventlog
from typing import Dict, List, Union

import win32security
import win32api

from constants.windows.constants import ConstRds
from utils.rds import (
    psh_request_handler,
    WMIHandler,
)

from models.rds import (
    RdsUserModel,
    RdsLogoutModel,
    RdsRemoteAppModel,
    RemoteAppUsersModel,
)


class RdsService:
    """
    Сервис взаимодействия с RDS.

    Handled errors:
    - OSAPIError (из модуля excepts): the error that occurs
      when executing a request API ОС.
    - AuthenticationError (из модуля excepts): the error that occurs
      during authentication.
    """

    def __init__(self, wmi_handler: WMIHandler, const: ConstRds):
        self._wmi_handler = wmi_handler
        self.GET_REMOTE_APP_USERS_CMD = const.GET_REMOTE_APP_USERS_CMD
        self.LOGOFF_USER_CMD = const.LOGOFF_USER_CMD

    def get_sessions(self) -> List[Dict[str, Union[str, int, bool]]]:
        """
        Function returning active rds sessions.
        """
        sessions = self._wmi_handler.get_sessions()
        return sessions

    def login_user(self, params: RdsUserModel) -> bool:
        """
        Authorization function in rds.
        """
        # LogonUser() return a handle to a token that represents the specified user.
        # Returned handle is a primary token that you can use in calls to the
        # CreateProcessAsUser() function.
        # If you specify the LOGON32_LOGON_NETWORK flag, LogonUser() returns
        # an impersonation token that you cannot use in CreateProcessAsUser()
        # unless you call DuplicateTokenEx() to convert it to a primary token.
        # LOGON32_PROVIDER_WINNT40: Use the NTLM logon provider.
        token = win32security.LogonUser(
            params.login,
            params.domain,
            params.password,
            win32security.LOGON32_LOGON_NETWORK,
            win32security.LOGON32_PROVIDER_WINNT40,
        )
        is_authenticated = bool(token)
        win32api.CloseHandle(token)
        eventlog.write_event("Auth is done.", "INFO")
        return is_authenticated

    def test_connect(self, params: RdsUserModel) -> bool:
        """
        Method that return response of trying to connect to rds
        """
        # TODO Пока не понятно как правильно делать проверку тестового подключения по RDP.
        # Оставим текущую логику закомментированной.

        # RDS_HOSTNAME = '192.168.3.13'
        # login = params.login
        # pwd = params.password

        # os.system(
        #     'cmd/c "cmdkey /generic:"{hostname}" /user:"{login}" \
        #         /pass:"{pwd}""'.format(hostname=RDS_HOSTNAME, login=login, pwd=pwd)
        #     )
        # os.system('cmd/c "mstsc /v:{hostname}"'.format(hostname=RDS_HOSTNAME))
        # os.system('cmd/c "cmdkey /delete:{hostname}"'.format(hostname=RDS_HOSTNAME))

        # Временное решение делать проверку посредством попытки авторизации на AD.
        result = self.login_user(params)
        return result

    def logout_user(self, params: RdsLogoutModel) -> bool:
        """
        Method disconnect user from RDS
        """
        sessions = self._wmi_handler.get_sessions()
        session_id = params.session_id
        # In future use host server
        # Example: logoff /VM <session id> /server:<host server>
        session_found = False
        for session in sessions:
            if session["session_id"] == session_id:
                session_found = True
                os.system(
                    'cmd/c "echo y|LOGOFF /VM {session_id}"'.format(session_id=session_id)
                )
                break

        if session_found:
            eventlog.write_event(f"Logout user {session_id}", "INFO")
        else:
            eventlog.write_event("User is not disconnected. Session ID not found", "ERROR")

        return session_found

    def get_collections(self) -> List[Dict[str, str]]:
        """
        Getting RDS collections.
        """
        collections = self._wmi_handler.get_collections()
        return collections

    def get_remote_app_users(self, params: RemoteAppUsersModel) -> List[str]:
        """
        Getting users that have access to published
        app by collection name and remote app alias.
        """
        collection = params.collection
        app_alias = params.app_alias
        # Getting users list using powershell handler. Todo: soon to be replaced with WMI handler.
        output = psh_request_handler(
            self.GET_REMOTE_APP_USERS_CMD.format(collection, app_alias)
        )  # FIXME[aap]: В текущей реализации невозможно понять, что команда завершилась с ошибкой
        app_users = [users.lstrip("UserGroups : ") for users in output]
        return app_users

    def get_remote_apps(self, params: RdsRemoteAppModel) -> List[Dict[str, str]]:
        """
        Getting published remote apps from RDS by collection name.
        """
        collection = params.collection
        apps = self._wmi_handler.get_remote_apps(collection)
        return apps

    def get_collection_info(self, collection_name: str) -> Dict[str, str]:
        """
        Getting collection info by collection name.
        """
        collection_info = self._wmi_handler.get_collection_info(collection_name)
        return collection_info
