import json
from typing import Any, Dict, List

import dbus
import eventlog
from excepts import OSAPIError

from constants.linux.constants import ConstStal


# FIXME: данная реализация предназначена для корректного выполнения тестов
try:
    from models.stal import StalLogoutModel, StalRemoteAppModel, StalUserModel
except ImportError:
    from session_agent.models.stal import StalLogoutModel, StalUserModel

from utils.base import extract_ip


class StalService:
    """
    Сервис взаимодействия со STAL.

    Handled errors:
    - OSAPIError (из модуля excepts): the error that occurs
      when executing a request API ОС.
    - AuthenticationError (из модуля excepts): the error that occurs
      during authentication.
    """

    __bus = None
    __stal = None
    __stal_iface = None

    # Class implemented via Singleton
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, "_instance"):
            cls._instance = super(StalService, cls).__new__(cls)
        return cls._instance

    # DI для проброса констант
    def __init__(self, const: ConstStal):
        self.const = const

        self.DEPTH = const.DEPTH
        self.DISPLAY_WIDTH = const.DISPLAY_WIDTH
        self.DISPLAY_HEIGHT = const.DISPLAY_HEIGHT
        self.SESSION_STATES = const.SESSION_STATES

    @property
    def stal_iface(self):
        if not self.__stal_iface:
            self.__bus = dbus.SystemBus()
            self.__stal = self.__bus.get_object("ru.uveon.stal", "/ru/uveon/stal")
            self.__stal_iface = dbus.Interface(self.__stal, "ru.uveon.stal")
            return self.__stal_iface

        return self.__stal_iface

    def version(self) -> str:
        """
        Method for gets STAL versions.
        """
        # Example: "STAL v.${MAJOR}.${MINOR}"
        version: dbus.String = self.stal_iface.stalTest()
        return version

    def status(self) -> bool:
        """
        Method for gets STAL status.
        """
        # dbus return 1 or True
        result = self.stal_iface.stalStatus()
        status = bool(result)
        return status

    def get_session_info(self, session_id: dbus.UInt32) -> dbus.String:
        """
        Gets session information.
        """
        session_info: dbus.String = self.stal_iface.stalSessionInfoJson(session_id)
        return session_info

    def __load_session_info(self, session_id: dbus.UInt32) -> None:
        """
        Records session information.
        """
        info: dbus.String = self.get_session_info(session_id)
        self.__session_info = json.loads(info)

    def __clear_session_info(self) -> None:
        """
        Removes session information.
        """
        self.__session_info = None

    def __get_session_state(self) -> str:
        """
        Get session state.
        """
        is_session_active = self.__session_info.get('remote:connected')
        if is_session_active is None:
            # FIXME:Для stal меньше v.1.1 проверяем наличие ключа 'pid3'
            is_session_active = self.__session_info["pid3"] != 0
        return self.SESSION_STATES[is_session_active]

    def __get_session_is_app(self) -> bool:
        """
        Get the value of whether the session is an application.
        Parameter value 'seamless:mode':
        if mode == false - terminal.
        if mode == true - app.
        """
        is_app_mode = self.__session_info["seamless:mode"]
        if isinstance(is_app_mode, str):  # special for STAL V0.3
            is_app_mode = is_app_mode == "true"
        return is_app_mode

    def __check_session_create(self, session_id: dbus.Int32) -> None:
        """Check session STAL creation."""
        if session_id == -1:
            raise OSAPIError("STAL session not created!")

    def get_sessions(self) -> List[Dict[str, Any]]:
        """
        Method for gets STAL sessions.
        """
        # Example:
        # dbus.Array(
        #     [
        #         dbus.Struct(
        #             (dbus.String("test1"), dbus.UInt32(59101)), signature=None
        #         )
        #     ],
        #     signature=dbus.Signature("(su)"),
        # )
        dbus_sessions: dbus.Array[dbus.Struct] = self.stal_iface.stalSessions()
        host_server = extract_ip()
        sessions = []
        for session in dbus_sessions:
            username, session_id = session
            self.__load_session_info(session_id)
            state = self.__get_session_state()
            is_desktop = not self.__get_session_is_app()
            self.__clear_session_info()
            sessions.append(
                {
                    "username": username,
                    "session_id": session_id,
                    "host_server": host_server,  # for logoff in future
                    "state": state,
                    "is_desktop": is_desktop,
                }
            )
        return sessions

    def login_user(self, params: StalUserModel) -> bool:
        """
        Method for start of a new user session.
        """
        is_success: dbus.Boolean = self.stal_iface.stalLogin(
            params.login,  # user
            params.domain,  # domain
            params.password,  # pass
        )
        eventlog.write_event("Auth is done.", "INFO")
        return bool(is_success)

    def disconnect_user(self, params: StalLogoutModel) -> bool:
        """
        Method for disconnect user.
        """
        uint32_session_id = dbus.UInt32(params.session_id)
        is_success: dbus.Boolean = self.stal_iface.stalDisconnect(
            uint32_session_id,
        )
        eventlog.write_event(f"User {uint32_session_id} disconnected", "INFO")
        return bool(is_success)

    def logout_user(self, params: StalLogoutModel) -> bool:
        """
        Method for shutdown user session.
        Session closes assync ~600ms.
        """
        is_success: dbus.Boolean = self.stal_iface.stalSessionShutdown(
            params.session_id,
        )
        eventlog.write_event(f"Logout user {params.session_id} ", "INFO")
        return bool(is_success)

    def test_connect(self, params: StalUserModel) -> bool:
        """
        Method for authentication.
        """
        is_success: dbus.Boolean = self.stal_iface.stalTestConnect(
            params.login,  # user
            params.domain,  # domain
            params.password,  # pass
        )
        return bool(is_success)

    def session_request(self, params: StalUserModel) -> int:
        """
        Method for creating user session in STAL.
        """
        # session_id может быть -1, если сессия не была создана
        session_id: dbus.Int32 = self.stal_iface.busSessionRequest(
            params.login,  # user
            params.password,  # pass
            self.DEPTH,  # depth
            self.DISPLAY_WIDTH,  # width
            self.DISPLAY_HEIGHT,  # height
        )
        self.__check_session_create(session_id)
        return session_id

    def program_request(self, params: StalRemoteAppModel) -> int:
        """
        Method for creating user session in STAL in single program mode.
        """
        session_id: dbus.Int32 = self.stal_iface.busProgramRequest(
            params.login,  # user
            params.domain,  # domain
            params.password,  # pass
            self.DISPLAY_WIDTH,  # width
            self.DISPLAY_HEIGHT,  # height
            params.program,  # program
        )
        self.__check_session_create(session_id)
        return session_id

    def get_remote_apps(self) -> List[str]:
        """Getting published remote apps from STAL.
        """
        remote_apps: dbus.Array = self.stal_iface.stalProgramList()
        return remote_apps
