import logging
import subprocess

import pythoncom
import wmi

from constants.windows.constants import ConstRds as const

logger = logging.getLogger(__name__)


def check_instance_is_rds() -> bool:
    """function returns the answer whether the server is rds.

    Handled errors:
    - OSAPIError (из модуля excepts): the error that occurs
      when executing a request API ОС

    Returns:
        response (dict): json response
    """
    result = False
    # Making dict "check_list" through dict comprehension to further
    # check the existence of rds processes
    check_list = {proc_name: False for proc_name in const.RDS_PROC_NAMES}

    try:
        import psutil

        for process in psutil.process_iter():
            if process.name() in const.RDS_PROC_NAMES:
                check_list.update({process.name(): True})
        result = all(check_list.values())
    finally:
        logger.debug(f"The following processes have been checked: {check_list}")

    return result


def win_cli_output_decoder(win_string: str) -> str:
    """function returns a string with normal Cyrillic characters.
    Functionality will be updated.

    Args:
        win_string (string): cp1251 encoded string

    Returns:
        response (string): cp886 encoded string
    """

    return win_string.encode("cp1251").decode("cp866")


def psh_request_handler(command: str) -> list:
    """Powershell output handler. Method will be replaced with WMI handler later"""
    output = (
        subprocess.run(
            ["powershell", "-Command", command],
            universal_newlines=True,
            stdout=subprocess.PIPE,
        )
        .stdout.strip()
        .splitlines()
    )

    return output


class WMIHandler:
    """This class represent a methods for retrieving
    TS data from Windows Server using WMI.
    """
    # WMI namespaces paths
    _remoteapp_namespace = "root\\cimv2\\rdms"
    _ts_namespace = "root\\cimv2\\TerminalServices"
    _cimv2_namespace = "root\\cimv2"

    @staticmethod
    def __get_wmi_handler(namespace: str):
        # Initialize the COM libraries for the calling thread.
        # Required for proper work with WMI providers.
        pythoncom.CoInitialize()
        return wmi.WMI(namespace=namespace)

    @classmethod
    def get_collections(cls) -> list:
        """
        Getting collections.

        Returns: dict of RDS collections aliases.
        """
        wmi_handler = cls.__get_wmi_handler(cls._ts_namespace)
        wmi_collections = wmi_handler.Win32_RDCentralPublishedRemoteDesktop()

        # Alias - Alias of the desktop.
        # ShowInPortal - True if this application should be shown in the TS Web Access; otherwise, False.
        collections = [
            {
                "name": collection.Alias,
                "is_desktop": collection.ShowInPortal,
            }
            for collection in wmi_collections
        ]

        return collections

    @classmethod
    def get_sessions(cls) -> list:
        """Getting active RDS sessions"""
        # Getting all desktop TS collections aliases
        desktop_collections = [
            collection['name'] for collection in cls.get_collections() if collection['is_desktop']
        ]

        # Getting all active sessions
        wmi_handler = cls.__get_wmi_handler(namespace=cls._cimv2_namespace)
        wmi_sessions = wmi_handler.Win32_SessionDirectorySessionEx()

        sessions = []

        for session in wmi_sessions:
            # Checking if session either in desktop collection or remote app.
            is_desktop = session.CollectionAlias in desktop_collections
            is_application = session.ApplicationType == const.REMOTE_APP_PROC_NAME
            # This check required to filter out direct RDP connections.
            if any([is_desktop, is_application]):
                sessions.append(
                    {
                        "username": session.UserName,
                        "session_id": session.SessionID,
                        "host_server": session.HostServer,  # for logoff in future
                        "state": const.SESSION_STATES[
                            const.SESSION_ID_STATES.get(session.SessionState)
                            # Сreated session has SessionState is STATE_ACTIVE,
                            # if reconnected to session has SessionState is STATE_CONNECTED
                            in (const.STATE_ACTIVE, const.STATE_CONNECTED)
                        ],
                        "is_desktop": is_desktop,
                    }
                )

        return sessions

    @classmethod
    def get_remote_apps(cls, collection: str) -> list:
        """Getting published remote apps using win32 class Win32_RDMSPublishedApplication.

        Args:
            - collection (str): RDS collection alias.

        Returns: list of published remote apps by their collection name.

        """
        remote_apps_info = []

        wmi_handler = cls.__get_wmi_handler(namespace=cls._remoteapp_namespace)
        remote_apps = wmi_handler.Win32_RDMSPublishedApplication(PoolName=collection)

        for app in remote_apps:
            remote_apps_info.append(
                {
                    "Alias": app.AppAlias,
                    "DisplayName": app.DisplayName,
                }
            )

        return remote_apps_info

    @classmethod
    def get_collection_info(cls, collection_name: str) -> dict:
        """Getting collection info by collection name

        Args:
            collection_name (str): contains collection name

        Returns:
            result (dict): json response
        """
        wmi_handler = cls.__get_wmi_handler(namespace=cls._ts_namespace)
        rds_collections = wmi_handler.Win32_RDCentralPublishedDeploymentSettings(
            PublishingFarm=collection_name
        )
        # Сheck that the collections are not empty and they have one element
        if rds_collections and len(rds_collections) == 1:
            [collection] = rds_collections
            return {"gateway_name": collection.GatewayName}
        return {"gateway_name": None}
