import logging

from cachetools import TLRUCache
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.types import CertificatePublicKeyTypes
from datetime import datetime, timedelta
from default import config_exist, ConfigHandler as config
from fastapi import Security
from fastapi.security import HTTPAuthorizationCredentials, HTTPBearer
# Fixme: these imports are made this way to make pyinstaller work.
#  Need to find another way for import migration in pyinstaller.
import jose
import jose.jwt as jwt
from typing import Dict
from excepts import AuthenticationError
from environment.checks import PlatformEnvironment


logger = logging.getLogger(__name__)


if PlatformEnvironment.is_linux():
    from constants.linux.constants import ConstLinuxEnv as const
if PlatformEnvironment.is_windows():
    from constants.windows.constants import ConstWindowsEnv as const

# Todo: при обращении к несуществующим секциям в конфиге, в логи пишется универсальная ошибка.
#  Необходимо реализовать возможность писать в логи персонализированную ошибку для каждой секции/параметра.
#  Это нужно для понимания зачем пользователю необходимы те или иные параметры в конфигурационном файле.
try:
    if config_exist():
        PRIVATE_KEY_PATH = config.select("SSL", "PRIVATE_KEY_PATH")
        CERTIFICATE_PATH = config.select("SSL", "CERTIFICATE_PATH")
    else:
        PRIVATE_KEY_PATH = const.SESSION_AGENT_CONFIG["SSL"]["PRIVATE_KEY_PATH"]
        CERTIFICATE_PATH = const.SESSION_AGENT_CONFIG["SSL"]["CERTIFICATE_PATH"]
except:
    logger.error("Для подписания токенов авторизации необходимо указать путь до сертификата и "
                 "приватного ключа. Пожалуйста, заполните данные параметры в конфигурационном "
                 "файле: PRIVATE_KEY_PATH, CERTIFICATE_PATH")


def get_public_key(path=CERTIFICATE_PATH) -> CertificatePublicKeyTypes:
    """
    Reads the certificate file.
    The certificate must exist.

    Returns:
        public_key (CertificatePublicKeyTypes): public key
    """
    with open(path, "rb") as file:
        certificate = file.read()
    cert_obj = x509.load_pem_x509_certificate(certificate, default_backend())
    public_key = cert_obj.public_key()
    return public_key


def get_private_key(path=PRIVATE_KEY_PATH) -> bytes:
    """
    Reads the private key file.
    The key must exist.

    Returns:
        private_key_data (bytes): private key
    """
    with open(path, "rb") as file:
        private_key_data = file.read()
    return private_key_data


def custom_ttu(_key, value, now):
    """
    ttu = time-to-use for TLRUCache.
    Checking the TOKEN_TITLE and assigning the right time for the TOKEN_TITLE.
    """
    if value == const.ACCESS_TOKEN_TITLE:
        return now + timedelta(seconds=const.ACCESS_TOKEN_TTL)
    elif value == const.REFRESH_TOKEN_TITLE:
        return now + timedelta(seconds=const.REFRESH_TOKEN_TTL)


class JWTAuthorization:
    """
    JWT authentication mechanism

    Handled errors:
    - AuthenticationError (module excepts): the error that occurs
    during authentication
    - JWTError: called in case of validation error jwt token
    """

    security = HTTPBearer()

    def __init__(self):
        self._private_key = None
        self._public_key = None
        self.algorithm = "RS256"
        self.token_whitelist = TLRUCache(maxsize=const.CACHE_LIMIT, ttu=custom_ttu, timer=datetime.now)

    @property
    def private_key(self) -> bytes:
        """
        The method retrieves the contents of the private key
        """

        if not self._private_key:
            self._private_key = get_private_key()
        return self._private_key

    @property
    def public_key(self) -> CertificatePublicKeyTypes:
        """
        The method retrieves the contents of certificate
        """
        if not self._public_key:
            self._public_key = get_public_key()
        return self._public_key

    def cache_verification(self, token) -> bool:
        """
        The method that checks the token in the allowed cache list
        """
        verify = token in self.token_whitelist  # Checking that the token has not been revoked
        return verify

    def refresh(self, refresh_token) -> Dict[str, str]:
        """
        A method that updates the refresh token
        and returns a new pair of access, refresh tokens
        """
        try:
            payload = jwt.decode(refresh_token, self.public_key,
                                 algorithms=[self.algorithm])

            token_type = payload.get('typ')
            token_is_valid = self.cache_verification(refresh_token)
            if token_type != const.REFRESH_TOKEN_TITLE or not token_is_valid:
                raise AuthenticationError("Invalid refresh token")
            # The used refresh token must be revoked
            del self.token_whitelist[refresh_token]
            tokens = self.create_tokens(payload)

            return tokens
        except jwt.ExpiredSignatureError:
            raise AuthenticationError("Expired refresh token")

        except jose.JWTError as ex:
            raise AuthenticationError(f"Invalid refresh token: {str(ex)}")

    def verify(self, auth: HTTPAuthorizationCredentials = Security(security)) -> dict:
        """
        The method that validates the token in the request handler
        """
        try:
            token = auth.credentials
            payload = jwt.decode(token, self.public_key,
                                 algorithms=[self.algorithm])
            token_is_valid = self.cache_verification(token)

            if not token_is_valid:
                raise AuthenticationError("Token is not valid")

        except jose.JWTError as ex:
            raise AuthenticationError(f"JWT validation error: {str(ex)}")

        return payload

    def generate_tokens(self, payload: dict, ttl: int, typ: str) -> str:
        """
        A method that provides generation of a new token
        including payload, lifetime and token type
        """
        release = datetime.utcnow()
        expire = release + timedelta(seconds=ttl)
        token_payload = payload.copy()
        token_payload.update({"exp": expire, "iat": release, "typ": typ})
        headers = {"alg": self.algorithm}
        token = jwt.encode(token_payload, self.private_key,
                           algorithm=self.algorithm, headers=headers)

        self.token_whitelist[token] = typ

        return token

    def create_tokens(self, payload: dict) -> Dict[str, str]:
        """
        Method for creating a new pair of access, refresh tokens
        """
        access_token = self.generate_tokens(payload, const.ACCESS_TOKEN_TTL,
                                            const.ACCESS_TOKEN_TITLE)
        refresh_token = self.generate_tokens(payload, const.REFRESH_TOKEN_TTL,
                                             const.REFRESH_TOKEN_TITLE)

        return dict(access_token=access_token, refresh_token=refresh_token)
